#include "stdafx.h"
#include <cstdarg>
#include <ctype.h>
#include "assembler.h"

#pragma warning(disable: 4996)

bool Assembler::Assemble(char* inputFilename, FILE* pInput, uint8vec* outputBuffer)
{
	m_inputFilename = inputFilename;
	m_outputBuffer = outputBuffer;
	m_assemblyFinished = false;

	char numberBuffer[10];

	for (m_lineNumber = 1; !m_assemblyFinished && fgets(m_inputLine, kInputLineLength-1, pInput); m_lineNumber++)
	{
		// initialize tokenization
		m_inputLineLength = strlen(m_inputLine);
		m_tokenEnd = -1;

		while (1)
		{
			// end of line?
			char* first = ReadToken();
			int firstLen = strlen(first);
			if (firstLen == 0)
				break;

			// can't support program binaries larger than 4K in size
                     if (GetCurrentOutputAddress() >= kAddressSpace)
			{
				ReportError("assembly address out of range");
				return false;
			}

			string lineLabel = "L_";
			lineLabel += inputFilename;
			lineLabel += "_";
			sprintf(numberBuffer, "%d", m_lineNumber);
			lineLabel += numberBuffer;

			m_labels[lineLabel] = GetCurrentOutputAddress();

			// is it a mnemonic?
			if (OpcodeEntry* pEntry = GetOpcodeTableEntry(first))
			{
				char* mnemo = first;
				char* operand = ReadToken();

				if (operand[0] == '\0')
				{
					if (!ParseImplied(pEntry, mnemo))
						return false;
				}
				else if (operand[0] == '#')
				{
					if (!ParseImmediate(pEntry, mnemo, operand))
						return false;
				}
				else
				{
					if (!ParseAddressed(pEntry, mnemo, operand))
						return false;
				}
			}
			// is it a label?
			else if (first[0] == '+' || first[0] == '-' || first[firstLen-1] == ':')
			{
				if (!ParseLabel(first))
					return false;
			}
                    // is it a define?
                    else if (stricmp(first, "#define") == 0)
                    {
				        char* name = ReadToken();
				        if (name[0] == '\0')
				        {
					        ReportError("missing #define constant name");
					        return false;
				        }
                        
				        char* value = ReadToken();
				        if (!ParseDefine(name, value))
					        return false;
                    }
                    // is it a directive?
                    else if (first[0] == '.')
                    {
				if (!ParseDirective(first))
					return false;
			}
			else
			{
				ReportError("unrecognized identifier '%s'", first);
				return false;
			}
		}
	}

	// report any unresolved references
	if (!m_pendingReferences.empty())
	{
		for (PendingReferenceList::iterator it = m_pendingReferences.begin(); it != m_pendingReferences.end(); ++it)
		{
			m_lineNumber = it->second[0].lineNumber; // report the line number of the first reference to this label
			ReportError("could not resolve label '%s'", it->first.c_str());
		}

		return false;
	}

	return true;
}

bool Assembler::ParseImplied(OpcodeEntry* pEntry, char* mnemo)
{
	uint8 opcode = pEntry->opcodes[kImplied];
	if (opcode != kInvalidOp)
	{
		m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | 0);
	}
	else
	{
		ReportError("implied addressing not supported for '%s'", mnemo);
		return false;
	}

	return true;
}

bool Assembler::ParseImmediate(OpcodeEntry* pEntry, char* mnemo, char* operand)
{
	uint8 opcode = pEntry->opcodes[kImmediate];
	if (opcode != kInvalidOp)
	{
		int constValue;
		
		if (operand[1] == '<')
		{
			constValue = GetOperandValue(&operand[1]) >> 4; // hi nibble
		}
		else if (operand[1] == '>')
		{
			constValue = GetOperandValue(&operand[1]) & 0xF; // low nibble
		}
		else
		{
			constValue = GetOperandValue(operand);
		}

		if (constValue > 15 || constValue < -15)
		{
			ReportError("immediate value '%s' exceeds size of nibble", &operand[1]);
			return false;
		}
		if (constValue < 0)
		{
			constValue += 16;
		}

		uint8 immValue = constValue;

		if (stricmp("_andi", mnemo) == 0)
		{
			m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | 0); // ANDI = NORI 0; NORI ~imm
			immValue = (~immValue) & 0xF;
		}

		m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | immValue);

		if (stricmp("_ori", mnemo) == 0)
			m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | 0); // ORI = NORI imm; NORI 0
	}
	else
	{
		ReportError("immediate addressing not supported for '%s'", mnemo);
		return false;
	}

	return true;
}

int Assembler::GetOperandValue(char* operand)
{
	int immValue;
	if (operand[1] == '$')
		immValue = strtol(&operand[2], NULL, 16); // hex constant
	else if (isdigit(operand[1]) || operand[1] == '-')
		immValue = strtol(&operand[1], NULL, 10); // decimal constant
	else if (strlen(operand) == 4 && operand[1] == '\'' && operand[3] == '\'')
		immValue = operand[2]; // character constant	
    else
    {
		string constantName(&operand[1]);
		LabelToAddressMap::iterator labelIt = m_labels.find(constantName);
		if (labelIt == m_labels.end())
		{
			ReportError("undefined constant '%s'", &operand[1]);
			return 0xFFFFFFFF;
		}
		immValue = m_labels[constantName];
    }

	return immValue;
}

bool Assembler::ParseAddressed(OpcodeEntry* pEntry, char* mnemo, char* operand)
{
    string operandStr = operand;

	int address;
	if (operand[0] == '$')
	{
		address = (int) strtoul(&operand[1], NULL, 16);
	}
	else if (isdigit(operand[0]))
	{
		address = (int) strtoul(&operand[0], NULL, 10);
	}
	else if (operandStr.find_first_of('+') != string::npos && operandStr.find_first_of('+') > 0)
       {
           // computed address
           unsigned int plusIndex = operandStr.find_first_of('+');
           string labelPart = operandStr.substr(0, plusIndex);
           string numberPart = (plusIndex < operandStr.length() - 1) ? operandStr.substr(plusIndex + 1) : "";
           
            // trim trailing spaces
            size_t endpos = labelPart.find_last_not_of(" \n\r\t");
            if( string::npos != endpos )
            {
                labelPart = labelPart.substr( 0, endpos+1 );
            }

            // trim leading spaces
            size_t startpos = numberPart.find_first_not_of(" \n\r\t");
            if( string::npos != startpos )
            {
                numberPart = numberPart.substr( startpos );
            }

            LabelToAddressMap::iterator labelIt = m_labels.find(labelPart.c_str());
	     if (labelIt != m_labels.end())
	     {
		  // existing label match
		  address = labelIt->second;
                if (!isdigit(numberPart.c_str()[0]))
                {
                    ReportError("computed address '%s' does not have a decimal offset", operand);
		      return false;
                }
                int offset = (int) strtoul(numberPart.c_str(), NULL, 10); // decimal offset
                address += offset;
	     }
	     else
	     {
                ReportError("computed address '%s' references a constant that's not yet defined", operand);
		  return false;
            }
       }
       else
	{
		LabelToAddressMap::iterator labelIt = m_labels.find(operand);
		if (labelIt != m_labels.end())
		{
			// existing label match
			address = labelIt->second;
		}
		else
		{
			// forward label reference
			m_pendingReferences[operand].push_back(PendingReference(m_lineNumber, GetCurrentOutputAddress()));
			address = 0;
		}
	}

	if (address >= kAddressSpace)
	{
		ReportError("operand address '%s' out of range", operand);
		return false;
	}

	uint8 opcode;
	int operandLen = strlen(operand);
	if (operandLen >= 3 && stricmp(&operand[operandLen-2], ",x") == 0)
	{
		// indexed
		opcode = pEntry->opcodes[kIndexed];
		if (opcode == kInvalidOp)
		{
			ReportError("indexed addressing not supported for '%s'", mnemo);
			return false;
		}
	}
	else
	{
		// absolute
		opcode = pEntry->opcodes[kAbsolute];
		if (opcode == kInvalidOp)
		{
			ReportError("absolute addressing not supported for '%s'", mnemo);
			return false;
		}
	}



	if (pEntry->size == 1)
	{
		if (address >= 16)
		{
			ReportError("address '%s' is outside data address space", operand);
			return false;
		}

		m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | (address));
	}
	else
	{
		if (address >= kAddressSpace)
		{
			ReportError("address '%s' is outside address space", operand);
			return false;
		}

		m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | (address >> 8));
		m_outputBuffer->push_back(address & 0xff);

		if (stricmp("_orm", mnemo) == 0)
			m_outputBuffer->push_back(0xE0); // ORM = NORM mem; NORI 0
	}

	return true;
}

bool Assembler::ParseLabel(char* label)
{
	// discard the trailing colon for named labels
	if (label[0] != '+' && label[0] != '-')
	{
		label[strlen(label)-1] = '\0';
	}
	
	string labelName(label);
	int labelAddress = GetCurrentOutputAddress();

	// if it's not a +ahead label, then save it
	if (label[0] != '+')
	{
		LabelToAddressMap::iterator labelIt = m_labels.find(labelName);
		if (labelIt != m_labels.end())
		{
			// if it's not a -back label, report an error
			if (label[0] != '-')
			{
				ReportError("label '%s' was already defined in this context", label);
				return false;
			}
		}

		m_labels[label] = labelAddress;
	}

	// if it's not a -back label, resolve any pending references to this label
	if (label[0] != '-')
	{
		PendingReferenceList::iterator referenceIt = m_pendingReferences.find(labelName);
		if (referenceIt != m_pendingReferences.end())
		{
			vector<PendingReference>& referenceAddresses = referenceIt->second;
			for (vector<PendingReference>::iterator it = referenceAddresses.begin(); it != referenceAddresses.end(); ++it)
			{
				int addressToFix = it->outputAddress;
				(*m_outputBuffer)[addressToFix] |= (labelAddress >> 8);
				(*m_outputBuffer)[addressToFix+1] = labelAddress & 0xff;
			}

			m_pendingReferences.erase(referenceIt);
		}
	}

	return true;
}

bool Assembler::ParseDefine(char* name, char* valueString)
{
    if (valueString[0] == '\0')
    {
        ReportError("missing #define value");
        return false;
    }

    string constantName(name);
    LabelToAddressMap::iterator labelIt = m_labels.find(constantName);
    if (labelIt != m_labels.end())
    {
        ReportError("constant '%s' was already defined in this context", name);
        return false;
    }

    int value;
    if (valueString[0] == '$')
    {
        value = (int) strtol(&valueString[1], NULL, 16);
    }
    else if (isdigit(valueString[0]))
    {
        value = (int) strtol(&valueString[0], NULL, 10);
    }
    else
    {
        ReportError("constant definition by another constant is unimplemnted");
        return false;
    }

    m_labels[name] = value;

    return true;
}


bool Assembler::ParseDirective(char* directive)
{
	if (stricmp(directive, ".end") == 0)
    {
		m_assemblyFinished = true;
		return true;
	}
    else if (stricmp(directive, ".org") == 0)
    {
        char* valueString = ReadToken();
        if (valueString[0] == '\0')
        {
            ReportError("missing .org address");
            return false;
        }

        int value;
        if (valueString[0] == '$')
        {
            value = (int) strtoul(&valueString[1], NULL, 16);
        }
        else if (isdigit(valueString[0]))
        {
            value = (int) strtoul(&valueString[0], NULL, 10);
        }
        else
        {
            ReportError(".org definition by a constant is unimplemnted");
            return false;
        }

        int currentAddress = GetCurrentOutputAddress();
        if (value < currentAddress)
        {
            ReportError(".org address %X is less than current address %X", value, currentAddress);
            return false;
        }

        for (int i=0; i<value-currentAddress; i++)
        {
            m_outputBuffer->push_back(0);
        }
    }
    else
    {
        ReportError("unrecognized directive '%s'", directive);
        return false;
    }

    return true;
}

void Assembler::ReportError(char* format, ...)
{
	va_list args;
    va_start(args, format);
	printf("%s(%d) : ERROR - ", m_inputFilename, m_lineNumber);
	vprintf(format, args);
	printf("\n");
	va_end(args);
}

Assembler::OpcodeEntry* Assembler::GetOpcodeTableEntry(const char* mnemo) const
{
	static OpcodeEntry opcodeTable[] = 
	{
		//instr size impl  abs   imm  indx
		{ "jmp", 2, 0xFF, 0x0C, 0xFF, 0xFF },
		{ "lit", 1, 0xFF, 0xFF, 0x04, 0xFF  },
		{ "in",  1, 0xFF, 0xFF, 0x05, 0xFF  },
		{ "ld",  2, 0xFF, 0x06, 0xFF, 0xFF  },
		{ "st",  2, 0xFF, 0x07, 0xFF, 0xFF  },
		{ "out", 1, 0xFF, 0xFF, 0x0D, 0xFF  },
		{ "jc",  2, 0xFF, 0x00, 0xFF, 0xFF  },
		{ "jnc", 2, 0xFF, 0x01, 0xFF, 0xFF  },
		{ "jz",  2, 0xFF, 0x08, 0xFF, 0xFF  },
		{ "jnz", 2, 0xFF, 0x09, 0xFF, 0xFF  },
		{ "addi", 1, 0xFF, 0xFF, 0x0A, 0xFF  },
		{ "addm", 2, 0xFF, 0x0B, 0xFF, 0xFF  },
		{ "nori", 1, 0xFF, 0xFF, 0x0E, 0xFF  },
		{ "norm", 2, 0xFF, 0x0F, 0xFF, 0xFF  },
		{ "cmpi", 1, 0xFF, 0xFF, 0x02, 0xFF  },
		{ "cmpm", 2, 0xFF, 0x03, 0xFF, 0xFF  },
		// aliases
		{ "jge",  2, 0xFF, 0x00, 0xFF, 0xFF  }, // JC
		{ "jl",   2, 0xFF, 0x01, 0xFF, 0xFF  }, // JNC
		// compound instructions
		{ "_ori",  1, 0xFF, 0xFF, 0x0E, 0xFF  }, // ORI imm: NORI imm; NORI 0
		{ "_orm",  2, 0xFF, 0x0F, 0xFF, 0xFF  }, // ORM mem: NORM mem; NORI 0
		{ "_andi", 1, 0xFF, 0xFF, 0x0E, 0xFF  }, // ANDI imm: NORI 0; NORI ~imm
	};

	for (int i=0; i < sizeof(opcodeTable)/sizeof(OpcodeEntry); i++)
	{
		if (stricmp(mnemo, opcodeTable[i].name) == 0)
			return &opcodeTable[i];
	}

	return NULL;
}

char* Assembler::ReadToken()
{
        bool inString = false;

	// already at end?
	if (m_tokenEnd+1 >= m_inputLineLength)
	{
		return &m_inputLine[m_inputLineLength];
	}

	// skip whitespace
	int nextTokenStart = m_tokenEnd + 1;
	while (isspace(m_inputLine[nextTokenStart]))
	{
		nextTokenStart++;
	}

	// anything left?
	if (nextTokenStart != m_inputLineLength)
	{
              if (m_inputLine[nextTokenStart] == '\'')
                  inString = true;

		// proceed to whitespace, comment, or EOL
		int nextTokenEnd = nextTokenStart+1;
		while (m_inputLine[nextTokenEnd] != '\0' && (inString || !isspace(m_inputLine[nextTokenEnd])) && m_inputLine[nextTokenEnd] != ';')
		{
                    if (m_inputLine[nextTokenEnd] == '\'')
                        inString = !inString;

		      nextTokenEnd++;
		}

		// mark end of token
		m_tokenEnd = nextTokenEnd;
		m_inputLine[m_tokenEnd] = '\0';
	}

	// if token is a comment, skip to EOL
	if (m_inputLine[nextTokenStart] == ';')
	{
		nextTokenStart = m_inputLineLength;
		m_tokenEnd = m_inputLineLength;
	}

	return &m_inputLine[nextTokenStart];
}