#include "stdafx.h"
#include "MachineState.h"
#include "LCD.h"

using namespace Simulator;
using namespace System::Globalization;
using namespace System::IO;
using namespace System::Windows::Forms;

MachineState::MachineState(String^ programRomPath, String^ microRomPath, unsigned long bootAddress) :
	m_programROM(gcnew array<unsigned char>(s_programRomSize)),
	m_RAM(gcnew array<unsigned char>(s_ramSize)),
	m_microrom(gcnew array<unsigned long>(s_microromSize)),
	m_lcd(gcnew LCD()),
	m_startAddress(bootAddress),
       m_buttonState(0xF)
{
	for (unsigned long addr = 0; addr<s_microromSize; addr++)
	{
		m_microrom[addr] = 0;
	}

	for (int i=0; i<2; i++)
	{
		if (!LoadMicrorom(microRomPath, i))
		{
			Application::Exit();
			return;
		}
	}

	if (!LoadProgramRom(programRomPath))
	{
		return;
	}

	Reset();
}

// Load a microrom image
bool MachineState::LoadMicrorom(String^ romPath, int romIndex)
{
	bool result = false;
	
	try 
	{
		FileStream^ stream = File::OpenRead(romPath + String::Format("\\microcode_{0}.bin", romIndex));
		BinaryReader^ br = gcnew BinaryReader(stream, Text::Encoding::ASCII);

		unsigned long addr = 0;
		array<Byte>^ byteArray;

		while (br->PeekChar() != -1) 
		{
			byteArray = br->ReadBytes(1);

			m_microrom[addr] |= byteArray[0] << (romIndex * 8);
			addr++;	
		}

		if (addr != 2048)
			throw gcnew ApplicationException("wrong microrom size");

		result = true;
	}
	catch (Exception^)
	{
		MessageBox::Show(String::Format("Could not load microcode_{0}.bin", romIndex));
	}

	return result;
}

bool MachineState::LoadProgramRom(String^ romPath)
{
	bool result = false;
	
	try 
	{
		FileStream^ stream = File::OpenRead(romPath );
		BinaryReader^ br = gcnew BinaryReader(stream, Text::Encoding::ASCII);

		unsigned long addr = 0;
		array<Byte>^ byteArray;

		while (br->PeekChar() != -1) 
		{
			byteArray = br->ReadBytes(1);

			m_programROM[addr] |= byteArray[0];
			addr++;	
		}

		if (addr != 4096)
		  MessageBox::Show("wrong program ROM size");

		result = true;
	}
	catch (Exception^)
	{
		MessageBox::Show(String::Format("Could not load program ROM {0}", romPath));
	}

	return result;
}

// Reset the machine to the initial state
void MachineState::Reset()
{
	Phase = 0;
	RegPC = m_startAddress;
	m_lcd->ClearLCD();
}

// Get a word of microinstruction data
unsigned long MachineState::GetMicroinstruction(unsigned long microAddress)
{
	return m_microrom[microAddress];
}

// Compute the current micro address
unsigned long MachineState::GetCurrentMicroaddress()
{
	unsigned long microAddress = GetCurrentInstruction() * 8 + Phase;

	if (FlagNotCarry)
	{
		microAddress += 4;
	}
	if (FlagNotZero)
	{
		microAddress += 2;
	}

	return microAddress;
}

// Get the current microinstruction data
unsigned long MachineState::GetCurrentMicroinstruction()
{
	return m_microrom[GetCurrentMicroaddress()];
}

unsigned char MachineState::GetCurrentInstruction()
{
	return (Fetch >> 4);
}

unsigned char MachineState::GetCurrentImmediate()
{
	return (Fetch & 0xF);
}

// Get the LCD text 
String^ MachineState::LCDText() 
{ 
	return m_lcd->LCDText(); 
}

// Extract the value of an individual field from a bitfield
unsigned long MachineState::FieldValue(unsigned long bitfield, BitPos pos, int len)
{
	// compute mask of 2^len - 1
	unsigned long mask = 1;
	for (int i=0; i<len; i++)
	{
		mask *= 2;
	}
	mask = mask - 1;

	return (bitfield >> ((int)pos - len + 1)) & mask;
}

// Simulate one clock cycle, executing one micro-instruction
unsigned long MachineState::MicroStep()
{
	unsigned long addressModified = 0xFF;

	unsigned long microOp = GetCurrentMicroinstruction();

	// Fetch
	if (Phase == 0)
	     Fetch = m_programROM[RegPC];

	// ALU
	bool notCarryOut, equalOut;
	unsigned char aluResult = RunALU(microOp, notCarryOut, equalOut);

	// assign flags
	if (IsBitZero(microOp, BitPos::kLoadFlags))
	{
		FlagNotCarry = notCarryOut;
		FlagNotZero = aluResult == 0 ? false : true;
	}

	// store ALU result to A
	if (IsBitZero(microOp, BitPos::kLoadA))
	{
		RegA = aluResult;
	}

	// data bus value
	unsigned char dbus = 0;
	int dbusCount = 0;
	if (IsBitZero(microOp, BitPos::kOEImm))
	{
		dbus = GetCurrentImmediate();
		dbusCount++;
	}
	if (IsBitZero(microOp, BitPos::kCSRAM) && IsBitOne(microOp, BitPos::kWERAM))
	{
		dbus = (m_RAM[GetCurrentMemoryAddress()] & 0xF);
		dbusCount++;
	}
	if (IsBitZero(microOp, BitPos::kOEIn))
	{
		dbus = m_buttonState;
		dbusCount++;
	}
	if (IsBitZero(microOp, BitPos::kOEALU))
	{
		dbus = aluResult & 0xF; 
		dbusCount++;
	}

	if (dbusCount > 1)
	{
		MessageBox::Show("too many data bus drivers");
	}

	// store data bus 
	if (IsBitZero(microOp, BitPos::kLoadOut))
	{
            // OUT registers 
            unsigned char imm = GetCurrentImmediate();
	     if ((imm & 0x1) == 0)
            {
                // high-to-low transition on E?
                if (((Out0 & 0x1) == 1) && 
                    ((dbus & 0x1) == 0))
                {
                    unsigned long RS = (Out0 & 0x2) >> 1;
                    m_lcd->WriteNibble(RS, Out1);
                }

                Out0 = dbus;
            }

            if ((imm & 0x2) == 0)
            {
                Out1 = dbus;
            }
	}
	if (IsBitZero(microOp, BitPos::kCSRAM) && IsBitZero(microOp, BitPos::kWERAM))
	{
		m_RAM[GetCurrentMemoryAddress()] = dbus;
		addressModified = GetCurrentImmediate();
	}

	// PC
	if (IsBitZero(microOp, BitPos::kNotLoadPC))
		RegPC = GetCurrentMemoryAddress();
	else if (IsBitOne(microOp, BitPos::kIncPC))
		RegPC++;
		
	// phase
	Phase = 1 - Phase;

	return addressModified;
}

// simulate the behavior of a 74LS181
unsigned char MachineState::RunALU(unsigned long microOp, bool% notCarryOut, bool% equalOut)
{
	unsigned char a;
	unsigned char b;
	unsigned char result;
	long tempResult;

	// left operand
	a = RegA;

	// right operand
	if (IsBitZero(microOp, BitPos::kOEImm))
	{
		b = GetCurrentImmediate();
	}
	if (IsBitZero(microOp, BitPos::kCSRAM) && IsBitOne(microOp, BitPos::kWERAM))
	{
		b = (m_RAM[GetCurrentMemoryAddress()] & 0xF);
	}
	if (IsBitZero(microOp, BitPos::kOEIn))
	{
		b = m_buttonState;
	}
	if (IsBitZero(microOp, BitPos::kOEALU))
	{
		b = 0; // error: combinatorial loop
	}

	unsigned char c = (unsigned char)FieldValue(microOp, BitPos::kNotCarryIn, 1);
	unsigned long function = FieldValue(microOp, BitPos::kS3, 4);

	if (IsBitOne(microOp, BitPos::kM))
	{
		// logical mode
		switch (function)
		{
		case 0:
			result = ~a;
			break;
		case 1:
			result = ~(a|b);
			break;
		case 2:
			result = (~a)&b;
			break;
		case 3:
			result = 0;
			break;
		case 4:
			result = ~(a&b);
			break;
		case 5:
			result = ~b;
			break;
		case 6:
			result = a^b;
			break;
		case 7:
			result = a^(~b);
			break;
		case 8:
			result = (~a)|b;
			break;
		case 9:
			result = ~(a^b);
			break;
		case 10:
			result = b;
			break;
		case 11:
			result = a&b;
			break;
		case 12:
			result = 0xFF;
			break;
		case 13:
			result = a|(~b);
			break;
		case 14:
			result = a|b;
			break;
		case 15:
			result = a;
			break;
		}

              result = (unsigned char)(result & 0xF);

		notCarryOut = false;
		equalOut = (result == 15);
	}
	else
	{
		// arithmetic mode
		switch (function)
		{
		case 0:
			tempResult = (long)a + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 1:
			tempResult = (long)(a|b) + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 2:
			tempResult = (long)(a|(~b)) + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 3:
			tempResult = -(long)c;
			notCarryOut = tempResult < 0;
			break;
		case 4:
			tempResult = (long)a + (a&(~b)) + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 5:
			tempResult = (long)(a|b) + (a&(~b)) + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 6:
			tempResult = (long)a - b - c;
			notCarryOut = tempResult < 0;
			break;
		case 7:
			tempResult = (long)(a&(~b)) - c;
			notCarryOut = tempResult < 0;
			break;
		case 8:
			tempResult = (long)a + (a&b) + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 9:
			tempResult = (long)a + b + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 10:
			tempResult = (long)(a|b) + (a&b) + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 11:
			tempResult = (long)(a&b) - c;
			notCarryOut = tempResult < 0;
			break;
		case 12:
			tempResult = (long)a + a + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 13:
			tempResult = (long)(a|b) + a + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 14:
			tempResult = (long)(a|(~b)) + a + 1 - c;
			notCarryOut = tempResult < 16;
			break;
		case 15:
			tempResult = (long)a - c;
			notCarryOut = tempResult < 0;
			break;
		}

		result = (unsigned char)(tempResult & 0xF);

		equalOut = (result == 15);
	}

	return (result & 0xF);
}