#include "stdafx.h"
#include "MainForm.h"
#include "MachineState.h"
#include "Disassembly.h"
#include "MicroDisassembly.h"
#include "InputDialog.h"

using namespace Simulator;
using namespace System::Windows::Forms;
using namespace System::IO;

// Initialize the simulator GUI
void MainForm::InitializeSimulation()
{
	m_tooltip->SetToolTip(m_continueButtton, "Run");
	m_tooltip->SetToolTip(m_stepButton, "Step Over");
	m_tooltip->SetToolTip(m_microStepButton, "Micro-Step");
	m_tooltip->SetToolTip(m_pauseButton, "Pause");
	m_tooltip->SetToolTip(m_resetButton, "Reset");

	try
	{
		m_bootImage = safe_cast<String^>(Application::UserAppDataRegistry->GetValue("Boot Image"));
		if (!m_bootImage)
		{
			m_bootImage = ".\\microchess.bin";
		}
	}
	catch(Exception^)
	{
		m_bootImage = ".\\microchess.bin";
	}

	try
	{
		String^ bootAddress = safe_cast<String^>(Application::UserAppDataRegistry->GetValue("Boot Address"));
		m_bootAddress = UInt32::Parse(bootAddress);
	}
	catch(Exception^)
	{
		m_bootAddress = 0;
	}

	try
	{
		m_refreshCycles = safe_cast<unsigned long>(Application::UserAppDataRegistry->GetValue("Refresh Cycles"));
	}
	catch(Exception^)
	{
		m_refreshCycles = 200000;
	}

	m_pauseButton->Enabled = false;
       m_statusLabel->Text = "Stopped"; 
       m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
	m_running = false;	
	m_stepsSinceLastInterrupt = 0;
	m_memoryViewAddress = 0;
	m_sourceDisassemblyStartAddress = 0xFFFFFFFF;
	m_memoryBreakpointAddress = 0xFFFFFF;
	m_machineState = gcnew MachineState(m_bootImage, ".", m_bootAddress);
	m_disassembler = gcnew Disassembly(m_machineState);
	m_microDisassembler = gcnew MicroDisassembly(m_machineState);

	m_currentInstructionAddress = m_machineState->RegPC;
	m_currentStackFrameAddress = m_currentInstructionAddress;

	// Set the initial condition codes so all the lights are off. In the real
	// hardware, they are uninitialized at startup.
	m_machineState->FlagNotCarry = false;
	m_machineState->FlagNotZero = false;

	array<String^>^ parts = m_bootImage->Split('\\');
	this->Text = "Nibbler - " + parts[parts->Length - 1];
	
	LoadSourceLabels(m_bootImage->Substring(0, m_bootImage->LastIndexOf('.')) + ".sym");

	m_stepOverDestinationAddress = 0xFFFFFF;
	m_steppingOut = false;
	m_stepOutCount = 0;
       m_clocksSinceReset = 0;

	m_stackFrames.Clear();

	RefreshData();
}

bool MainForm::RetireInstruction()
{
	bool forceBreak = false;

	// next instruction
	m_executingOpcode = m_machineState->GetCurrentInstruction();
	m_executingPC = m_machineState->RegPC;

	return forceBreak;
}

// handle the application idle event, to do background processing
void MainForm::OnIdle(System::Object^  sender, System::EventArgs^  e)
{
	if (m_running)
	{
		TimeSpan frameTime = DateTime::Now - m_lastUpdateTime;
		unsigned int numInstructionsToRun = m_refreshCycles; 
              unsigned int clockCount = 0;

		while (numInstructionsToRun)
		{
			bool hitMicroBreakpoint = false;
			bool forceBreak = false;

			do
			{
				unsigned long addressModified = m_machineState->MicroStep();
                            clockCount++;
				if (addressModified == m_memoryBreakpointAddress)
				{
					MessageBox::Show(String::Format("Memory breakpoint {0:X1}", addressModified));
					forceBreak = true;
					break;
				}

				// microbreakpoint hit?
				if (CurrentMicroaddressIsAtBreakpoint())
				{
					hitMicroBreakpoint = true;
					break;
				}
			} while (m_machineState->Phase != 0);

			if (m_machineState->Phase == 0)
			{
				m_currentInstructionAddress = m_machineState->RegPC;
				m_currentStackFrameAddress = m_currentInstructionAddress;
				forceBreak = RetireInstruction();
			}

			// breakpoint?
			if (hitMicroBreakpoint ||
				m_sourceBreakpoints.ContainsKey(m_currentInstructionAddress) ||
				m_currentInstructionAddress == m_stepOverDestinationAddress ||
				forceBreak)
			{
				m_running = false;
				m_pauseButton->Enabled = false;
				m_continueButtton->Enabled = true;
				m_stepButton->Enabled = true;
				m_microStepButton->Enabled = true;
				m_clearBreakpointsButton->Enabled = true;
				m_stepOverDestinationAddress = 0xFFFFFF;
                            m_statusLabel->Text = "Breakpoint hit"; 
                            m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
				
				break;
			}

			numInstructionsToRun--;
		}

              m_clocksSinceReset += clockCount;
              // tick: equal to 100 nanoseconds or one ten-millionth of a second. There are 10,000 ticks in a millisecond. 
              m_statusLabel->Text = String::Format("Running: {0:f2} MHz", (double)20*m_refreshCycles / m_lastFrameDuration.Ticks); 
              m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);

		// TODO: the simulation actually seems to run more slowly if RefreshData() isn't called
		// I think the Application::Idle event is generated less frequently if the UI doesn't change
		RefreshData();
	}

       DateTime now = DateTime::Now;
       m_lastFrameDuration = now - m_lastUpdateTime;
	m_lastUpdateTime = now;
}

// Test if the current microinstruction address matches a micro-source breakpoint
bool MainForm::CurrentMicroaddressIsAtBreakpoint()
{
	unsigned long microaddress = m_machineState->GetCurrentMicroaddress();
	unsigned long microProgramBaseAddress = microaddress >> 5;
	
	if (m_microSourceBreakpoints.ContainsKey(microProgramBaseAddress))
	{
		List<unsigned long>^ conditionCodes = m_microSourceBreakpoints[microProgramBaseAddress];
		for each (unsigned long conditionCode in conditionCodes)
		{
			if ((conditionCode & (1 << ((microaddress >> 1) & 0x1))) && // E
				(conditionCode & (1 << (((microaddress >> 2) & 0x1) + 2)))) // C
			{
				return true;
			}
		}

		return false;
	}
	else
	{
		return false;
	}
}

// refresh the UI with the machine state data
void MainForm::RefreshData()
{
	// 4-bit registers
	m_aLabel->Text = String::Format("{0:X1}", m_machineState->RegA);

	// address registers
	m_pcLabel->Text = String::Format("{0:X3}", m_machineState->RegPC);

	// control
	m_fetchLabel->Text = String::Format("{0:X2}", m_machineState->Fetch); 
	m_phaseLabel->Text = String::Format("{0:X1}", m_machineState->Phase);
       m_out0Label->Text = String::Format("{0:X1}", m_machineState->Out0);
       m_out1Label->Text = String::Format("{0:X1}", m_machineState->Out1);
	m_opLabel->Text = m_disassembler->GetOpcodeDescription(m_machineState->GetExecutingInstruction());

	// flags
	m_cRadioButton->Checked = !m_machineState->FlagNotCarry;
	m_eRadioButton->Checked = !m_machineState->FlagNotZero;

	// memory
	FillMemoryTextBox(m_memoryViewTextBox, m_memoryViewAddress, 0, false);

	// disassembly
	RefreshDisassemblyView(true);
	RefreshMicroDisassemblyView();

	// LCD
	m_lcdTextBox->Text = m_machineState->LCDText();

	m_microDisassemblyViewTitle->Text = m_disassembler->GetOpcodeDescription(m_machineState->GetExecutingInstruction()) + " - Micro-Disassembly";
}

void MainForm::FillConsoleTextBox(System::Windows::Forms::TextBox^ textBox)
{
}

// Refresh the source disassembly view
void MainForm::RefreshDisassemblyView(bool scrollToCurrentInstruction)
{
	unsigned long newSourceDisassemblyStartAddress = Math::Max(0,(m_currentStackFrameAddress-50)/50*50);
	if (newSourceDisassemblyStartAddress == m_sourceDisassemblyStartAddress)
	{
		m_disassembler->UpdateDisassemblyListView(m_sourceListView, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, scrollToCurrentInstruction);
	}
	else
	{
		m_sourceDisassemblyStartAddress = newSourceDisassemblyStartAddress;
		m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);
	}
}

// Refresh the micro source disassembly view
void MainForm::RefreshMicroDisassemblyView()
{
	m_microDisassembler->FillDisassemblyListView(m_microSourceListView, %m_microSourceBreakpoints);
}

// Fill a memory view textbox
void MainForm::FillMemoryTextBox(TextBox^ textBox, unsigned long startAddress, unsigned long highlightAddress, bool highlight)
{
	const int linesToShow = 128;

	String^ result;

       startAddress = 0;

	for (int line=0; line<linesToShow; line++)
	{
		result += String::Format("{0:X3}", startAddress + line*16);
		for (int byte=0; byte<16; byte++)
		{
			if (highlight && startAddress + line*16 + byte == highlightAddress)
			{
				result += ">";
			}
			else
			{
				result += " ";
			}

			unsigned char data = m_machineState->PeekRAM(startAddress + line*16 + byte);
			result += String::Format("{0:X1}", data);
		}

		if (line < linesToShow-1)
		{
			result += "\r\n";
		}
	}

	textBox->Text = result;
}

// Handle a click on the micro-step button
void MainForm::m_microStepButton_Click(System::Object^  sender, System::EventArgs^  e) 
{
	m_machineState->MicroStep();
       m_clocksSinceReset++;

	if (m_machineState->Phase == 0)
	{
		m_currentInstructionAddress = m_machineState->RegPC;
		RetireInstruction();
	}

	m_currentStackFrameAddress = m_currentInstructionAddress;
       m_statusLabel->Text = "Stopped"; 
       m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
	RefreshData();
}

void MainForm::m_stepButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	do
	{
		m_machineState->MicroStep();
              m_clocksSinceReset++;
	} while (m_machineState->Phase != 0);


	m_currentInstructionAddress = m_machineState->RegPC;
	m_currentStackFrameAddress = m_currentInstructionAddress;

	RetireInstruction();
       m_statusLabel->Text = "Stopped"; 
       m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
	RefreshData();
}

// Handle a click on the reset button
void MainForm::m_resetButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_machineState->Reset();
	m_pauseButton->Enabled = false;
	m_continueButtton->Enabled = true;
	m_stepButton->Enabled = true;
	m_microStepButton->Enabled = true;
	m_clearBreakpointsButton->Enabled = true;
	m_running = false;
	m_currentInstructionAddress = m_machineState->RegPC;
	m_currentStackFrameAddress = m_currentInstructionAddress;
	m_stepOverDestinationAddress = 0xFFFFFF;
	m_steppingOut = false;
	m_stepOutCount = 0;
	m_executingOpcode = 0xFF;
       m_clocksSinceReset = 0;
       m_statusLabel->Text ="Stopped"; 
       m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
	m_stackFrames.Clear();

	RefreshData();
}

// Handle a click on the Go button
void MainForm::m_continueButtton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_pauseButton->Enabled = true;
	m_continueButtton->Enabled = false;
	m_stepButton->Enabled = false;
	m_microStepButton->Enabled = false;
	m_clearBreakpointsButton->Enabled = false;
	m_statusLabel->Text = "Running"; 
       m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
	
	m_running = true;
}

// Handle a click on the pause button
void MainForm::m_pauseButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_continueButtton->Enabled = true;
	m_continueButtton->Focus();
	m_pauseButton->Enabled = false;
	m_stepButton->Enabled = true;
	m_microStepButton->Enabled = true;
	m_clearBreakpointsButton->Enabled = true;
	m_running = false;
	m_steppingOut = false;
	m_stepOutCount = 0;

	m_stepOverDestinationAddress = 0xFFFFFF;
	m_statusLabel->Text = "Stopped"; 
       m_timeLabel->Text = String::Format("Time: {0:N1} us", m_clocksSinceReset/s_clockSpeedMHz);
	RefreshData();
}

// Handle the "Open Program ROM" menu item
void MainForm::openProgramROMToolStripMenuItem_Click(System::Object^  sender, System::EventArgs^  e)
{
	OpenFileDialog^ dialog = gcnew OpenFileDialog();
	dialog->Title = "Select Program File";
	dialog->Filter = "bin files (*.bin)|*.bin";
	dialog->RestoreDirectory = true;

	if (dialog->ShowDialog() == System::Windows::Forms::DialogResult::OK)
	{
		m_bootImage = dialog->FileName;
		Application::UserAppDataRegistry->SetValue("Boot Image", m_bootImage);
		InitializeSimulation();
		m_machineState = gcnew MachineState(dialog->FileName, ".", m_bootAddress);
		RefreshData();
		RefreshDisassemblyView(true);
		//m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, %m_sourceBreakpoints, %m_stackFrames, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);
	}
}

// Handle the "Open Source Labels" menu item
void MainForm::toolStripMenuItem2_Click(System::Object^  sender, System::EventArgs^  e)
{
	OpenFileDialog^ dialog = gcnew OpenFileDialog();
	dialog->Title = "Select Symbol File";
	dialog->Filter = "symbol files (*.sym)|*.sym|label files (*.lbl)|*.lbl";
	dialog->RestoreDirectory = true;

	if (dialog->ShowDialog() == System::Windows::Forms::DialogResult::OK)
	{
		LoadSourceLabels(dialog->FileName);
	}
}

// Load the label data for source debugging
void MainForm::LoadSourceLabels(String^ filename)
{
	m_addressToSource.Clear();
	m_sourceStartAddress.Clear();

	// try lbl file if sym doesn't exist, or vice-versa
	if (!File::Exists(filename))
	{
		if (filename->EndsWith(".sym"))
		{
			filename = filename->Substring(0,filename->LastIndexOf(".")) + ".lbl";
		}
		else if (filename->EndsWith(".lbl"))
		{
			filename = filename->Substring(0,filename->LastIndexOf(".")) + ".sym";
		}
	}

	bool isCC65 = filename->EndsWith(".lbl");

	String^ dir = ".\\";

	int lastSlash = filename->LastIndexOf('\\');
	if (lastSlash >= 0)
	{
		dir = filename->Substring(0, lastSlash+1);
	}

	try
	{
		StreamReader^ sr = gcnew StreamReader(filename);

		try
		{
			String^ line;
		 
			// Read and display lines from the file until the end of 
			// the file is reached.
			while ( (line = sr->ReadLine()) != nullptr )
			{
				if (line->StartsWith("L_"))
				{
					int comment = line->IndexOf(';');
					line = line->Substring(0, comment);

					array<String^>^ parts = line->Split('=');
					String^ labelName = parts[0]->Substring(2, parts[0]->Length-3);
					unsigned long sourceAddress = UInt32::Parse(parts[1]->Substring(1), NumberStyles::HexNumber);

					int idx = labelName->LastIndexOf('_');
					String^ sourceFilename = labelName->Substring(0, idx);

					int sourceLineNumber = Int32::Parse(labelName->Substring(idx+1));

					idx = sourceFilename->LastIndexOf('_');
					if (idx >= 0)
					{
						sourceFilename = sourceFilename->Substring(0, idx) + "." + sourceFilename->Substring(idx+1);
					}
					sourceFilename = dir + sourceFilename;

					// addresses should be mapped to the highest line number possible in a file,
					// but to the file with the lowest line number
					FileAndLine fileAndLine;
					if (m_addressToSource.TryGetValue(sourceAddress, fileAndLine))
					{
						if (!fileAndLine.filename)
						{
							fileAndLine.filename = sourceFilename;
							fileAndLine.lineNumber = sourceLineNumber;
							m_addressToSource[sourceAddress] = fileAndLine;
						}
						else if (fileAndLine.filename->Equals(sourceFilename))
						{
							// same file, choose highest line
							if (sourceLineNumber > fileAndLine.lineNumber)
							{
								fileAndLine.lineNumber = sourceLineNumber;
								m_addressToSource[sourceAddress] = fileAndLine;
							}
						}
						else
						{
							// different file, choose lowest line
							if (sourceLineNumber < fileAndLine.lineNumber)
							{
								fileAndLine.filename = sourceFilename;
								fileAndLine.lineNumber = sourceLineNumber;
								m_addressToSource[sourceAddress] = fileAndLine;
							}
						}
					}
					else
					{
						// insert new entry for this address
						fileAndLine.filename = sourceFilename;
						fileAndLine.lineNumber = sourceLineNumber;
						fileAndLine.sourceLabel = "";
						m_addressToSource.Add(sourceAddress, fileAndLine);
					}
					
					unsigned long sourceStartAddress;
					if (m_sourceStartAddress.TryGetValue(sourceFilename, sourceStartAddress))
					{
						if (sourceAddress < sourceStartAddress)
						{
							m_sourceStartAddress[sourceFilename] = sourceAddress;
						}
					}
					else
					{
						// insert a new entry for the start address
						m_sourceStartAddress[sourceFilename] = sourceAddress;
					}
				}

				else
				{
					FileAndLine fileAndLine;
					unsigned long sourceAddress;

					if (isCC65)
					{
						array<String^>^ parts = line->Split('.');
						sourceAddress = UInt32::Parse(parts[0]->Substring(3,6), NumberStyles::HexNumber);
								
						m_addressToSource.TryGetValue(sourceAddress, fileAndLine);
						
						// update entry
						fileAndLine.sourceLabel = parts[1]->Trim();
						fileAndLine.filename = "unknown";
						fileAndLine.lineNumber = 1;
						m_addressToSource[sourceAddress] = fileAndLine;
					}
					else
					{
						int comment = line->IndexOf(';');
						if (comment >= 0)
						{
							line = line->Substring(0, comment);
						}

						array<String^>^ parts = line->Split('=');
						sourceAddress = UInt32::Parse(parts[1]->Substring(1), NumberStyles::HexNumber);
						
						m_addressToSource.TryGetValue(sourceAddress, fileAndLine);
						
						// update entry
						fileAndLine.sourceLabel = parts[0]->Trim();
						m_addressToSource[sourceAddress] = fileAndLine;
					}
				}
			}

			m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);			
		}
		finally
		{
			if ( sr )
			delete (IDisposable^)sr;	
		}
	}
	catch ( Exception^ ) 
	{
		MessageBox::Show( String::Format("Couldn't open symbol file {0}", filename), "File Error" );
	}
}

void MainForm::m_ShowDisassemblyCheckbox_CheckedChanged(System::Object^  sender, System::EventArgs^  e) 
{
	m_ignoreSource = m_ShowDisassemblyCheckbox->Checked;
	m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);
}

void MainForm::toolStripMenuItem4_Click(System::Object^  sender, System::EventArgs^  e)
{
	InputDialog^ inputDialog = gcnew InputDialog();
	inputDialog->Text = "Set Program Start Address";
	inputDialog->SetPrompt("Start Address (hex)");
	inputDialog->SetInputBoxText(String::Format("{0:X6}", m_bootAddress));

	System::Windows::Forms::DialogResult result = inputDialog->ShowDialog();
	if (result == System::Windows::Forms::DialogResult::OK)
	{
		m_bootAddress = System::Convert::ToUInt32(inputDialog->GetInputBoxText(), 16);
		Application::UserAppDataRegistry->SetValue("Boot Address", m_bootAddress);
	}
}

void MainForm::toolStripMenuItem5_Click(System::Object^  sender, System::EventArgs^  e)
{
InputDialog^ inputDialog = gcnew InputDialog();
	inputDialog->Text = "Set Memory Breakpoint";
	inputDialog->SetPrompt("Breakpoint Address (hex)");
	inputDialog->SetInputBoxText(String::Format("{0:X6}", m_memoryBreakpointAddress));

	System::Windows::Forms::DialogResult result = inputDialog->ShowDialog();
	if (result == System::Windows::Forms::DialogResult::OK)
	{
		m_memoryBreakpointAddress = System::Convert::ToUInt32(inputDialog->GetInputBoxText(), 16);
	}
}

void MainForm::toolStripMenuItem6_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_memoryBreakpointAddress = 0xFFFFFF;
}

void MainForm::m_buttonLeft_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(0, true);
}

void MainForm::m_buttonLeft_MouseUp(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(0, false);
}

void MainForm::m_buttonRight_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(1, true);
}

void MainForm::m_buttonRight_MouseUp(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(1, false);
}

void MainForm::m_buttonDown_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(2, true);
}

void MainForm::m_buttonDown_MouseUp(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(2, false);
}

void MainForm::m_buttonUp_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(3, true);
}

void MainForm:: m_buttonUp_MouseUp(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e)
{
    m_machineState->SetButton(3, false);
}